##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Support for non-versioned data embedded in versioned objects.

$Id$
"""

from .interfaces import INonVersionedData
from .VersionSupport import isAVersionableResource
from Acquisition import aq_base
from OFS.ObjectManager import ObjectManager
from zope.interface import implementer

try:
    # Optional support for references.
    from Products.References.Proxy import proxyBase
    from Products.References.PathReference import PathReference
except ImportError:
    isProxyOrReference = None
else:
    def isProxyOrReference(obj):
        if proxyBase(obj) is not aq_base(obj):
            return 1
        if isinstance(obj, PathReference):
            return 1
        return 0


def getNonVersionedDataAdapter(obj):
    """Returns an INonVersionedData adapter for any object.

    This is a super-simplistic adapter implementation.
    """
    base = aq_base(obj)
    # If the object implements INonVersionedData, let it say
    # what its items are.
    if INonVersionedData.providedBy(base):
        return obj
    # If the object is an ObjectManager, use the ObjectManager adapter.
    if isinstance(base, ObjectManager):
        return ObjectManagerNonVersionedDataAdapter(obj)
    # Otherwise use the standard adapter.
    return StandardNonVersionedDataAdapter(obj)


def listNonVersionedObjects(obj):
    return getNonVersionedDataAdapter(obj).listNonVersionedObjects()

def getNonVersionedData(obj):
    return getNonVersionedDataAdapter(obj).getNonVersionedData()

def removeNonVersionedData(obj):
    getNonVersionedDataAdapter(obj).removeNonVersionedData()

def restoreNonVersionedData(obj, dict):
    getNonVersionedDataAdapter(obj).restoreNonVersionedData(dict)



@implementer(INonVersionedData)
class StandardNonVersionedDataAdapter:
    """Non-versioned data adapter for arbitrary things.
    """

    def __init__(self, obj):
        self.obj = obj
        # __vc_ignore__, if set, is a tuple of attribute names to
        # manage independently of version control.
        self.attrs = getattr(obj, "__vc_ignore__", ())

    def listNonVersionedObjects(self):
        # Assume it's OK to clone all of the attributes.
        # They will be removed later by removeNonVersionedData.
        return ()

    def removeNonVersionedData(self):
        for attr in self.attrs:
            try:
                delattr(aq_base(self.obj), attr)
            except (AttributeError, KeyError):
                pass

    def getNonVersionedData(self):
        data = {}
        for attr in self.attrs:
            if hasattr( aq_base(self.obj), attr ):
                data[attr] = aq_base(getattr(aq_base(self.obj), attr))
        return data

    def restoreNonVersionedData(self, data):
        for attr in self.attrs:
            if attr in data:
                setattr(aq_base(self.obj), attr, data[attr])


@implementer(INonVersionedData)
class ObjectManagerNonVersionedDataAdapter(StandardNonVersionedDataAdapter):
    """Non-versioned data adapter for object managers.
    """

    def listNonVersionedObjects(self):
        contents = self.getNonVersionedData()['contents']
        return contents.values()

    def removeNonVersionedData(self):
        StandardNonVersionedDataAdapter.removeNonVersionedData(self)
        obj = self.obj
        removed = {}
        contents = self.getNonVersionedData()['contents']
        for name, value in contents.items():
            obj._delOb(name)
            removed[name] = 1
        if obj._objects:
            obj._objects = tuple([info for info in obj._objects
                                  if info['id'] not in removed])

    def getNonVersionedData(self):
        contents = {}
        attributes = StandardNonVersionedDataAdapter.getNonVersionedData(self)
        for name, value in self.obj.objectItems():
            if not isAVersionableResource(value):
                # This object should include the state of subobjects
                # that won't be versioned independently.
                continue
            if isProxyOrReference is not None:
                if isProxyOrReference(value):
                    # This object should include the state of
                    # subobjects that are references.
                    continue
            contents[name] = aq_base(value)
        return {'contents': contents, 'attributes': attributes}

    def restoreNonVersionedData(self, data):
        StandardNonVersionedDataAdapter.restoreNonVersionedData(
            self, data['attributes'])
        # First build "ignore", a dictionary that lists which
        # items were stored in the repository.
        # Don't restore over those.
        obj = self.obj
        ignore = {}
        for name in obj.objectIds():
            ignore[name] = 1
        # Restore the items of the container.
        for name, value in data['contents'].items():
            if name not in ignore:
                obj._setOb(name, aq_base(value))
                if not hasattr(obj, '_tree'):
                    # Avoid generating events, since nothing was ever really
                    # removed or added.
                    obj._objects += ({'meta_type': value.meta_type,
                                      'id': name},)
                # If there is a _tree attribute, it's very likely
                # a BTreeFolder2, which doesn't need or want the
                # _objects attribute.
                # XXX This is a hackish way to check for BTreeFolder2s.



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

__version__='$Revision: 1.11 $'[11:-2]

import time

from Acquisition import Implicit, aq_base
from AccessControl.class_init import InitializeClass
from Persistence import Persistent
from AccessControl import ClassSecurityInfo
from OFS.SimpleItem import SimpleItem
from ZODB._compat import BytesIO
from ZODB._compat import Pickler, Unpickler
from .nonversioned import listNonVersionedObjects, removeNonVersionedData


def cloneByPickle(obj, ignore_list=()):
    """Makes a copy of a ZODB object, loading ghosts as needed.

    Ignores specified objects along the way, replacing them with None
    in the copy.
    """
    ignore_dict = {}
    for o in ignore_list:
        ignore_dict[id(o)] = o

    def persistent_id(ob, ignore_dict=ignore_dict):
        if id(ob) in ignore_dict:
            return 'ignored'
        if getattr(ob, '_p_changed', 0) is None:
            ob._p_changed = 0
        return None

    def persistent_load(ref):
        assert ref == 'ignored'
        # Return a placeholder object that will be replaced by
        # removeNonVersionedData().
        placeholder = SimpleItem()
        placeholder.id = "ignored_subobject"
        return placeholder

    stream = BytesIO()
    p = Pickler(stream, 1)
    p.persistent_id = persistent_id
    p.dump(obj)
    stream.seek(0)
    u = Unpickler(stream)
    u.persistent_load = persistent_load
    return u.load()


class Version(Implicit, Persistent):
    """A Version is a resource that contains a copy of a particular state
       (content and dead properties) of a version-controlled resource.  A
       version is created by checking in a checked-out resource. The state
       of a version of a version-controlled resource never changes."""

    def __init__(self, version_id, obj):
        self.id = version_id
        self.date_created = time.time()
        self._data = None

    # These attributes are set by the createVersion method of the version
    # history at the time the version is created. The branch is the name
    # of the branch on which the version was created. The prev attribute
    # is the version id of the predecessor to this version. The next attr
    # is a sequence of version ids of the successors to this version.
    branch = 'mainline'
    prev = None
    next = ()

    security = ClassSecurityInfo()

    security.declarePublic('getId')
    def getId(self):
        return self.id

    security.declarePrivate('saveState')
    def saveState(self, obj):
        """Save the state of object as the state for this version of
           a version-controlled resource."""
        self._data = self.stateCopy(obj, self)

    security.declarePrivate('copyState')
    def copyState(self):
        """Return an independent deep copy of the state of the version."""
        data = self.__dict__.get('_data')  # Avoid __of__ hooks
        return self.stateCopy(data, self)

    security.declarePrivate('stateCopy')
    def stateCopy(self, obj, container):
        """Get a deep copy of the state of an object.

        Breaks any database identity references.
        """
        ignore = listNonVersionedObjects(obj)
        res = cloneByPickle(aq_base(obj), ignore)
        removeNonVersionedData(res)
        return res


InitializeClass(Version)



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

import os
import time

from AccessControl import getSecurityManager
from App.Common import package_home
from AccessControl.class_init import InitializeClass
from Persistence import Persistent
from ZODB.TimeStamp import TimeStamp

try:
    from ZODB.serialize import referencesf
except ImportError:  # < Zope 2.8 / ZODB 3.3
    from ZODB.referencesf import referencesf

_dtmldir = os.path.join( package_home( globals() ), 'dtml' )

use_vc_permission = 'Use version control'


def isAVersionableResource(obj):
    """ True if an object is versionable.

    To qualify, the object must be persistent (have its own db record), and
    must not have an true attribute named '__non_versionable__'."""

    if getattr(obj, '__non_versionable__', 0):
        return 0
    return hasattr(obj, '_p_oid')

class VersionInfo(Persistent):
    """A VersionInfo object contains bookkeeping information for version
       controlled objects. The bookkeeping information can be read (but
       not changed) by restricted code."""

    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, history_id, version_id, status):
        self.timestamp = time.time()
        self.history_id = history_id
        self.version_id = version_id
        self.status = status
        self.user_id = _findUserId()

    sticky = None

    CHECKED_OUT = 0
    CHECKED_IN = 1

    def branchName(self):
        if self.sticky is not None and self.sticky[0] == 'B':
            return self.sticky[1]
        return 'mainline'

    def clone(self, clear_sticky=0):
        info = VersionInfo(self.history_id, self.version_id, self.status)
        dict = info.__dict__
        for name, value in self.__dict__.items():
            dict[name] = value
        if clear_sticky:
            if 'sticky' in dict:
                del dict['sticky']
        info.user_id = _findUserId()
        info.timestamp = time.time()
        return info

InitializeClass(VersionInfo)



class ReadOnlyJar:
    """A read-only ZODB connection-like object that prevents changes."""

    def __init__(self, base):
        self.__base__ = base

    _invalidating = []

    def __getattr__(self, name):
        return getattr(self.__base__, name)

    def commit(*args, **kw):
        raise VersionControlError(
            'Old versions of objects cannot be modified.'
            )

    def abort(*args, **kw):
        pass



class VersionControlError(Exception):
    pass



def _findUserId():
    user = getSecurityManager().getUser()
    return user.getUserName()

def _findPath(object):
    path = object.getPhysicalPath()
    return '/'.join(path)

def _findModificationTime(object):
    """Find the last modification time for a version-controlled object.
       The modification time reflects the latest modification time of
       the object or any of its persistent subobjects that are not
       themselves version-controlled objects. Note that this will
       return None if the object has no modification time."""

    mtime = getattr(object, '_p_mtime', None)
    if mtime is None:
        return None

    latest = mtime
    conn = object._p_jar
    load = conn._storage.load
    refs = referencesf

    oids = [object._p_oid]
    done_oids = set()
    done = done_oids.__contains__
    first = 1

    while oids:
        oid = oids[0]
        del oids[0]
        if done(oid):
            continue
        done_oids.add(oid)
        try:
            p, serial = load(oid)
        except Exception:
            pass  # invalid reference!
        else:
            if first is not None:
                first = None
            else:
                if p.find(b'U\x0b__vc_info__') == -1:
                    mtime = TimeStamp(serial).timeTime()
                    if mtime > latest:
                        latest = mtime
            refs(p, oids)

    return latest



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
""" Test the ZVC machinery

$Id$
"""
import unittest

from .common import get_transaction
from .common import common_setUp
from .common import common_tearDown
from .common import common_commit

class VersionControlTests(unittest.TestCase):

    setUp = common_setUp
    tearDown = common_tearDown
    commit = common_commit

    do_commits = 0

    def testIsVersionableResource(self):
        # Test checking whether an object is a versionable resource.
        repository = self.repository
        document = self.document1
        nonversionable = self.document_nonversion
        self.assertTrue(repository.isAVersionableResource(document))
        self.assertFalse(repository.isAVersionableResource(nonversionable))
        self.assertFalse(repository.isAVersionableResource('foobar'))
        self.assertFalse(repository.isAVersionableResource(self))


    def testIsUnderVersionControl(self):
        # Test checking whether an object is under version control.
        repository = self.repository
        document = self.document1
        self.assertFalse(repository.isUnderVersionControl(document))
        repository.applyVersionControl(document)
        self.commit()
        self.assertTrue(repository.isUnderVersionControl(document))


    def testIsResourceUpToDate(self):
        # Test checking whether a versioned resource is up to date.
        repository = self.repository
        document = repository.applyVersionControl(self.document1)
        self.commit()
        self.assertTrue(repository.isResourceUpToDate(document))

        info = repository.getVersionInfo(document)
        first_version = info.version_id

        document = repository.checkoutResource(document)
        self.commit()
        document = repository.checkinResource(document, '')
        self.commit()

        document = repository.updateResource(document, first_version)
        self.commit()
        self.assertFalse(repository.isResourceUpToDate(document))

        document = repository.updateResource(document, None)
        self.commit()
        self.assertTrue(repository.isResourceUpToDate(document))


    def testIsResourceChanged(self):
        # Test checking whether a versioned resource has changed.
        repository = self.repository
        document = repository.applyVersionControl(self.document1)
        self.commit()
        if self.do_commits:
            self.assertFalse(repository.isResourceChanged(document))

        document = repository.checkoutResource(document)
        self.commit()
        if self.do_commits:
            self.assertFalse(repository.isResourceChanged(document))

        document.manage_edit('change 1', '')
        self.commit()
        if self.do_commits:
            self.assertTrue(repository.isResourceChanged(document))

        document = repository.checkinResource(document, '')
        self.commit()
        if self.do_commits:
            self.assertFalse(repository.isResourceChanged(document))


    def testVersionBookkeeping(self):
        # Check the consistency of the version bookkeeping info.
        repository = self.repository

        document = repository.applyVersionControl(self.document1)
        self.commit()

        info = repository.getVersionInfo(document)
        self.assertTrue(info.user_id == 'UnitTester')
        self.assertTrue(info.status == info.CHECKED_IN)
        self.assertTrue(info.sticky == None)
        first_version = info.version_id

        document = repository.checkoutResource(document)
        self.commit()
        info = repository.getVersionInfo(document)
        self.assertTrue(info.status == info.CHECKED_OUT)

        document = repository.checkinResource(document, '')
        self.commit()
        info = repository.getVersionInfo(document)
        self.assertTrue(info.status == info.CHECKED_IN)

        document = repository.updateResource(document, first_version)
        self.commit()
        info = repository.getVersionInfo(document)
        self.assertTrue(info.status == info.CHECKED_IN)
        self.assertTrue(info.version_id == first_version)

        branch_name = 'Bug Fix Branch'
        repository.makeActivity(document, branch_name)
        self.commit()
        document = repository.updateResource(document, branch_name)
        self.commit()
        info = repository.getVersionInfo(document)
        self.assertTrue(info.sticky == ('B', branch_name))

        document = repository.updateResource(document, 'mainline')
        self.commit()
        info = repository.getVersionInfo(document)
        self.assertTrue(info.sticky == None)

        document = repository.checkoutResource(document)
        self.commit()
        info = repository.getVersionInfo(document)
        self.assertTrue(info.status == info.CHECKED_OUT)

        document.manage_edit('change 1', '')
        self.commit()

        document = repository.uncheckoutResource(document)
        self.commit()
        info = repository.getVersionInfo(document)
        self.assertTrue(info.status == info.CHECKED_IN)


    def testApplyVersionControl(self):
        # Test checking whether a versioned resource is up to date.
        from Products.ZopeVersionControl.Utility import VersionControlError
        repository = self.repository

        document = repository.applyVersionControl(self.document1)
        self.commit()
        self.assertTrue(repository.isUnderVersionControl(document))

        # Make sure we can't do it a second time.
        self.assertRaises(VersionControlError,
                              repository.applyVersionControl,
                              document
                              )

        info = repository.getVersionInfo(document)

        # Check that the last log entry record is what we expect.
        record = repository.getLogEntries(document)[0]
        self.assertTrue(record.version_id == info.version_id)
        self.assertTrue(record.user_id == 'UnitTester')
        self.assertTrue(record.action == record.ACTION_CHECKIN)
        self.assertTrue(record.path == '/folder1/folder2/document1')


    def testCheckoutResource(self):
        # Test checking out a version controlled resource.
        from Products.ZopeVersionControl.Utility import VersionControlError
        repository = self.repository

        document = repository.applyVersionControl(self.document1)
        self.commit()

        info = repository.getVersionInfo(document)
        first_version = info.version_id

        repository.checkoutResource(document)
        self.commit()

        # Make sure you can't checkout a checked-out resource.
        self.assertRaises(VersionControlError,
                              repository.checkoutResource,
                              document
                              )

        # Check that the last log entry record is what we expect.
        record = repository.getLogEntries(document)[0]
        self.assertTrue(record.version_id == first_version)
        self.assertTrue(record.user_id == 'UnitTester')
        self.assertTrue(record.action == record.ACTION_CHECKOUT)
        self.assertTrue(record.path == '/folder1/folder2/document1')

        document = repository.checkinResource(document, '')
        self.commit()
        document = repository.updateResource(document, first_version)
        self.commit()

        # Make sure you can't checkout a non-up-to-date resource.
        self.assertRaises(VersionControlError,
                              repository.checkoutResource,
                              document
                              )

        # Check that the last log entry record is what we expect.
        record = repository.getLogEntries(document)[0]
        self.assertTrue(record.version_id == first_version)
        self.assertTrue(record.user_id == 'UnitTester')
        self.assertTrue(record.action == record.ACTION_UPDATE)
        self.assertTrue(record.path == '/folder1/folder2/document1')


    def testCheckinResource(self):
        # Test checking in a version controlled resource.
        from Products.ZopeVersionControl.Utility import VersionControlError
        repository = self.repository

        document = repository.applyVersionControl(self.document1)
        self.commit()

        # Make sure you can't checkin a checked-in resource.
        self.assertRaises(VersionControlError,
                              repository.checkinResource,
                              document, ''
                              )

        info = repository.getVersionInfo(document)
        first_version = info.version_id

        repository.checkoutResource(document)
        self.commit()

        document = repository.checkinResource(document, '')
        self.commit()

        # Check that the last log entry record is what we expect.
        record = repository.getLogEntries(document)[0]
        info = repository.getVersionInfo(document)
        self.assertTrue(record.version_id == info.version_id)
        self.assertTrue(record.user_id == 'UnitTester')
        self.assertTrue(record.action == record.ACTION_CHECKIN)
        self.assertTrue(record.path == '/folder1/folder2/document1')

        self.assertRaises(VersionControlError,
                              repository.checkinResource,
                              document, ''
                              )

        document = repository.updateResource(document, first_version)
        self.commit()

        self.assertRaises(VersionControlError,
                              repository.checkinResource,
                              document, ''
                              )


    def testUncheckoutResource(self):
        # Test uncheckout of a version controlled resource.
        from Products.ZopeVersionControl.Utility import VersionControlError
        repository = self.repository

        document = repository.applyVersionControl(self.document1)
        self.commit()

        info = repository.getVersionInfo(document)
        first_version = info.version_id

        repository.checkoutResource(document)
        self.commit()

        document.manage_edit('change 1', '')

        document = repository.uncheckoutResource(document)
        self.commit()

        info = repository.getVersionInfo(document)
        self.assertTrue(info.status == info.CHECKED_IN)

        # Check that the last log entry record is what we expect.
        record = repository.getLogEntries(document)[0]
        self.assertTrue(record.version_id == first_version)
        self.assertTrue(record.user_id == 'UnitTester')
        self.assertTrue(record.action == record.ACTION_UNCHECKOUT)
        self.assertTrue(record.path == '/folder1/folder2/document1')


    def testUpdateResource(self):
        # Test updating a version controlled resource.
        from Products.ZopeVersionControl.Utility import VersionControlError
        repository = self.repository

        document = repository.applyVersionControl(self.document1)
        self.commit()

        info = repository.getVersionInfo(document)
        first_version = info.version_id

        repository.labelResource(document, 'First Version', 1)

        repository.checkoutResource(document)
        self.commit()

        document = repository.checkinResource(document, '')
        self.commit()

        document = repository.updateResource(document, first_version)
        self.commit()
        info = repository.getVersionInfo(document)
        self.assertTrue(info.version_id == first_version)

        document = repository.updateResource(document, None)
        self.commit()

        document = repository.updateResource(document, 'First Version')
        self.commit()
        info = repository.getVersionInfo(document)
        self.assertTrue(info.version_id == first_version)

        # Check that the last log entry record is what we expect.
        record = repository.getLogEntries(document)[0]
        self.assertTrue(record.version_id == first_version)
        self.assertTrue(record.user_id == 'UnitTester')
        self.assertTrue(record.action == record.ACTION_UPDATE)
        self.assertTrue(record.path == '/folder1/folder2/document1')


    def testLabelResource(self):
        # Test labeling a version controlled resource.
        from Products.ZopeVersionControl.Utility import VersionControlError
        repository = self.repository

        document = repository.applyVersionControl(self.document1)
        self.commit()

        info = repository.getVersionInfo(document)
        first_version = info.version_id

        repository.labelResource(document, 'First Version', 1)

        document = repository.checkoutResource(document)
        self.commit()

        self.assertRaises(VersionControlError,
                              repository.labelResource,
                              document, 'First Version', 1
                              )

        document = repository.uncheckoutResource(document)
        self.commit()

        repository.makeActivity(document, 'Activity 1')

        self.assertRaises(VersionControlError,
                              repository.labelResource,
                              document, 'mainline', 1
                              )

        self.assertRaises(VersionControlError,
                              repository.labelResource,
                              document, 'Activity 1', 1
                              )

        document = repository.checkoutResource(document)
        self.commit()

        document = repository.checkinResource(document, '')
        self.commit()

        self.assertRaises(VersionControlError,
                              repository.labelResource,
                              document, 'First Version', 0
                              )

        document = repository.updateResource(document, 'First Version')
        self.commit()
        info = repository.getVersionInfo(document)
        self.assertTrue(info.version_id == first_version)


    def testActivityAPI(self):
        from Products.ZopeVersionControl.Utility import VersionControlError

        repository = self.repository
        document = self.document1

        repository.applyVersionControl(document)
        self.commit()

        self.assertTrue(repository.isUnderVersionControl(document))
        self.assertTrue(repository.isResourceUpToDate(document))
        if self.do_commits:
            self.assertFalse(repository.isResourceChanged(document))

        info = repository.getVersionInfo(document)
        self.assertTrue(info.sticky == None)
        first_version = info.version_id

        activity_name = 'My Big Project'

        repository.makeActivity(document, activity_name)
        self.commit()

        # Make sure we can't do it again for the same activity id.
        self.assertRaises(VersionControlError,
                              repository.makeActivity,
                              document, activity_name
                              )

        document = repository.updateResource(document, activity_name)
        self.commit()

        self.assertTrue(repository.isResourceUpToDate(document))
        if self.do_commits:
            self.assertFalse(repository.isResourceChanged(document))

        info = repository.getVersionInfo(document)
        self.assertTrue(info.version_id == first_version)
        self.assertTrue(info.sticky == ('B', activity_name))

        repository.checkoutResource(document)
        self.commit()

        document.manage_edit('activity change 1', '')
        self.commit()

        document = repository.checkinResource(document, 'activity change 1')
        self.commit()

        info = repository.getVersionInfo(document)
        self.assertTrue(info.sticky == ('B', activity_name))

        for n in range(10):
            activity_name = 'Activity %d' % n

            repository.makeActivity(document, activity_name)
            self.commit()

            info = repository.getVersionInfo(document)
            root_version = info.version_id

            document = repository.updateResource(document, activity_name)
            self.commit()

            self.assertTrue(repository.isResourceUpToDate(document))

            info = repository.getVersionInfo(document)
            self.assertTrue(info.sticky == ('B', activity_name))
            self.assertTrue(info.version_id == root_version)

            repository.checkoutResource(document)
            self.commit()

            document.manage_edit('activity change %d' % n, '')
            self.commit()

            document = repository.checkinResource(
                document, 'activity change %d' % n
                )
            self.commit()

            info = repository.getVersionInfo(document)
            self.assertTrue(info.sticky == ('B', activity_name))
            self.assertFalse(info.version_id == root_version)

        document = repository.updateResource(document, root_version)
        self.commit()

        self.assertFalse(repository.isResourceUpToDate(document))

        document = repository.updateResource(document, first_version)
        self.commit()

#        self.assertTrue(repository.isResourceUpToDate(document))




    def testSelectionByDate(self):
        # Test selection of versions by date.
        from Products.ZopeVersionControl.Utility import VersionControlError
        from DateTime.DateTime import DateTime
        import time

        repository = self.repository
        document = repository.applyVersionControl(self.document1)
        self.commit()

        document = repository.checkoutResource(document)
        self.commit()

        document.manage_edit('change 1', '')
        self.commit()

        document = repository.checkinResource(document, 'change 1')
        self.commit()

        repository.labelResource(document, 'change 1')
        self.commit()

        info = repository.getVersionInfo(document)
        first_version = info.version_id

        # Trickery: we'll hack the timestamp of the first version so
        # that we can act like it was created yesterday :)
        history = repository.getVersionHistory(info.history_id)
        version = history.getVersionById(info.version_id)
        orig_time = version.date_created
        new_time = orig_time - 86400.0
        timestamp = int(orig_time / 60.0)
        new_stamp = int(new_time / 60.0)
        version.date_created = new_time
        branch = history._branches['mainline']
        key, val = branch.m_order.items()[0]
        del branch.m_date[timestamp]
        branch.m_date[new_stamp] = key
        self.commit()

        for n in range(10):
            document = repository.checkoutResource(document)
            self.commit()
            change_no = 'change %d' % (n + 2)
            document.manage_edit(change_no, '')
            self.commit()
            document = repository.checkinResource(document, change_no)
            self.commit()
            repository.labelResource(document, change_no)
            self.commit()

        target_time = time.time() - 14400.0
        document = repository.updateResource(document, target_time)
        self.commit()
        info = repository.getVersionInfo(document)
        self.assertTrue(info.version_id == first_version)

        # Now do some branching and make sure we backtrack correctly
        # through the branch lineage when doing date selection.
        document = repository.updateResource(document, 'change 5')
        self.commit()

        for n in range(10):
            activity_name = 'Activity %d' % n
            repository.makeActivity(document, activity_name)
            self.commit()
            document = repository.updateResource(document, activity_name)
            self.commit()
            for i in range(10):
                repository.checkoutResource(document)
                self.commit()
                document.manage_edit('activity change %d' % i, '')
                self.commit()
                document = repository.checkinResource(document, '')
                self.commit()

        document = repository.updateResource(document, target_time)
        self.commit()
        info = repository.getVersionInfo(document)
        self.assertTrue(info.version_id == first_version)


    def testSelectionByLabel(self):
        # Test labeling and selection of versions using labels.
        from Products.ZopeVersionControl.Utility import VersionControlError

        repository = self.repository
        document = repository.applyVersionControl(self.document1)

        # Ensure that labeling and lookup by label works as expected.
        label_set = []
        for n in range(10):
            change_no = 'change %d' % n
            document = repository.checkoutResource(document)
            document.manage_edit(change_no, '')
            document = repository.checkinResource(document, change_no)
            repository.labelResource(document, change_no)
            info = repository.getVersionInfo(document)
            label_set.append((change_no, info.version_id))

        for label, version_id in label_set:
            document = repository.updateResource(document, label)
            info = repository.getVersionInfo(document)
            self.assertTrue(info.version_id == version_id)
            self.assertTrue(info.sticky == ('L', label))

        # Ensure that label moving works as expected and that we get an
        # error if we try to reuse a label without forcing a label move.
        document = repository.updateResource(document, 'change 0')
        repository.labelResource(document, 'change 1', force=1)
        document = repository.updateResource(document, 'change 1')
        info = repository.getVersionInfo(document)
        self.assertTrue(info.version_id == label_set[0][1])

        self.assertRaises(VersionControlError,
                              repository.labelResource,
                              document, 'change 3'
                              )


    def testGetVersionOfResource(self):
        # Test retrieving specific versions of resources.
        from Products.ZopeVersionControl.Utility import VersionControlError
        repository = self.repository

        document = repository.applyVersionControl(self.document1)
        self.commit()

        info = repository.getVersionInfo(document)
        history_id = info.history_id
        first_version = info.version_id

        repository.labelResource(document, 'First Version')

        for n in range(10):
            repository.checkoutResource(document)
            self.commit()
            document = repository.checkinResource(document, '')
            self.commit()

        # Make sure the "get version of resource" api is working.
        doc_copy = repository.getVersionOfResource(
            info.history_id, first_version
            )
        info = repository.getVersionInfo(doc_copy)
        self.assertTrue(info.version_id == first_version)
        self.assertTrue(info.sticky == ('V', first_version))
        self.assertTrue(document._p_oid != doc_copy._p_oid)
        self.assertTrue(document is not doc_copy)

        doc_copy = repository.getVersionOfResource(
            info.history_id, "First Version"
            )
        info = repository.getVersionInfo(doc_copy)
        self.assertTrue(info.version_id == first_version)
        self.assertTrue(info.sticky == ('L', 'First Version'))
        self.assertTrue(document._p_oid != doc_copy._p_oid)
        self.assertTrue(document is not doc_copy)


    def testDetectPersistentSubObjectChange(self):
        # Test detection of changes to persistent sub-objects.
        repository = self.repository
        folder2 = self.folder2
        document = self.document1

        # Place a resource w/ persistent subobjects under version control.
        repository.applyVersionControl(folder2)
        get_transaction().commit()

        document1 = getattr(folder2, 'document1')
        document1.manage_edit('spam spam', '')
        get_transaction().commit()

        self.assertTrue(repository.isResourceChanged(folder2))


    def testContainerVersioning(self):
        from OFS.DTMLDocument import addDTMLDocument
        # Verify that containers and items are versioned independently.
        repository = self.repository
        folder1 = self.app.folder1
        folder2 = folder1.folder2
        folder1.testattr = 'container_v1'
        folder2.testattr = 'item_v1'

        self.assert_(not repository.isUnderVersionControl(folder1))
        repository.applyVersionControl(folder1)
        folder1 = self.app.folder1
        self.assert_(repository.isUnderVersionControl(folder1))
        self.assert_(not repository.isUnderVersionControl(folder2))
        repository.applyVersionControl(folder2)
        folder2 = folder1.folder2
        self.assert_(repository.isUnderVersionControl(folder2))
        self.assert_(not repository.isUnderVersionControl(folder2.document1))

        # Make the first version of folder1 and check it in.
        repository.checkoutResource(folder1)
        folder1 = self.app.folder1
        repository.checkinResource(folder1)
        folder1 = self.app.folder1
        folder2 = folder1.folder2
        info = repository.getVersionInfo(folder1)
        first_version = info.version_id

        # Change folder1 and check it in again
        repository.checkoutResource(folder1)
        folder1 = self.app.folder1
        folder1.testattr = 'container_v2'
        addDTMLDocument(folder1, 'document3', file='some more text')
        repository.checkinResource(folder1)
        folder1 = self.app.folder1
        folder2 = folder1.folder2

        # Change folder2
        repository.checkoutResource(folder2)
        folder2 = folder1.folder2
        folder2.testattr = 'item_v2'

        # Now revert folder1 and verify that folder2 was not reverted.
        repository.updateResource(folder1, first_version)
        folder1 = self.app.folder1
        folder2 = folder1.folder2
        self.assertEqual(folder1.testattr, 'container_v1')
        self.assertEqual(folder2.testattr, 'item_v2')

        # Verify that document3 remains an item of the reverted folder1.
        self.assert_(hasattr(folder1, 'document3'))
        self.assert_(str(folder1.document3) == 'some more text')

        # Remove document3 and verify that it doesn't reappear upon revert.
        folder1._delObject('document3')
        repository.updateResource(folder1, '')
        folder1 = self.app.folder1
        self.assertEqual(folder1.testattr, 'container_v2')
        self.assertEqual(folder1.folder2.testattr, 'item_v2')
        self.assert_(not hasattr(folder1, 'document3'))


    def testNonVersionedAttribute(self):
        # Test a non-version-controlled attribute mixed with
        # a version-controlled attribute.
        self.document1.extra_attr = 'v1'
        self.document1.__vc_ignore__ = ('__ac_local_roles__',)
        self.document1.__ac_local_roles__ = {'sam': ['Manager',]}
        repository = self.repository
        document = repository.applyVersionControl(self.document1)
        info = repository.getVersionInfo(document)
        first_version = info.version_id
        repository.checkoutResource(document)
        self.document1.extra_attr = 'v2'
        self.document1.__ac_local_roles__ = {}
        repository.checkinResource(document)
        repository.updateResource(document, first_version)
        self.assertEqual(document.extra_attr, 'v1')
        self.assertEqual(document.__ac_local_roles__, {})

    def testNonVersionedAttributeWithAcquisition(self):
        # Test a non-version-controlled attribute
        # that is acquired does not cause an error
        self.document1.__vc_ignore__ = ('dummy_attr',)
        self.folder1.dummy_attr = 'dummy_attr'
        self.assertEqual( self.folder1.dummy_attr, self.document1.dummy_attr )

        repository = self.repository
        document = repository.applyVersionControl(self.document1)
        info = repository.getVersionInfo(document)
        first_version = info.version_id
        repository.checkoutResource(document)
        self.document1.extra_attr = 'new'
        repository.checkinResource(document)
        repository.updateResource(document, first_version)

class VersionControlTestsWithCommits(VersionControlTests):
    """Version control test suite with transaction commits that mimic
       the transaction commits that you would get with Web based usage."""
    do_commits = 1

def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(VersionControlTests))
    suite.addTest(unittest.makeSuite(VersionControlTestsWithCommits))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')




from setuptools import setup, find_packages

__version__ = '1.1.4'

setup(
    name='Products.ZopeVersionControl',
    version=__version__,
    description="Zope Version Control",
    long_description=(open('README.rst').read() + "\n" +
                      open('CHANGES.rst').read()),
    classifiers=[
        'Framework :: Zope2',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    license='ZPL',
    author='Zope Foundation and Contributors',
    author_email='zope-dev@zope.org',
    url='https://pypi.org/project/Products.ZopeVersionControl',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    namespace_packages=['Products'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'setuptools',
        'zope.interface',
        'Acquisition',
        'DateTime',
        'transaction',
        'six',
        'ZODB',
        'Zope2',
    ],
    entry_points={
        'zodbupdate.decode': [
            'decodes = Products.ZopeVersionControl:zodbupdate_decode_dict',
        ],
    },
)



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

from . import VersionHistory
from AccessControl.class_init import InitializeClass
from App.special_dtml import DTMLFile

import AccessControl
import OFS

# BBB Zope 2.12
try:
    from OFS.role import RoleManager
except ImportError:
    from AccessControl.Role import RoleManager


class ZopeVersionHistory(
    VersionHistory.VersionHistory,
    RoleManager,
    OFS.SimpleItem.Item,
    ):
    """The ZopeVersionHistory build on the core VersionHistory class to
       provide the Zope management interface and other product trappings."""

    security = AccessControl.ClassSecurityInfo()
    security.setDefaultAccess('deny')

    meta_type = 'Version History'

    manage_options=(
        ( {'label': 'Contents',    'action':'manage_main',
           'help': ('ZopeVersionControl', 'VersionHistory-Manage.stx')},
          {'label': 'Properties', 'action':'manage_properties_form',
           'help': ('ZopeVersionControl', 'VersionHistory-Properties.stx')},
        ) +
        RoleManager.manage_options +
        OFS.SimpleItem.Item.manage_options
        )

    icon='misc_/ZopeVersionControl/VersionHistory.gif'

    security.declareProtected('View management screens', 'manage_main')
    manage_main = DTMLFile('dtml/VersionHistoryManageMain', globals())
    manage_main._setName('manage_main')
    manage = manage_main

    security.declareProtected(
        'View management screens', 'manage_properties_form'
        )
    manage_properties_form = DTMLFile(
        'dtml/VersionHistoryProperties', globals()
        )

    security.declareProtected('Manage repositories', 'manage_edit')
    def manage_edit(self, REQUEST=None):
        """Change object properties."""
        if REQUEST is not None:
            message="Saved changes."
            return self.manage_properties_form(
                self, REQUEST, manage_tabs_message=message
                )

    def __getitem__(self, name):
        activity = self._branches.get(name)
        if activity is not None:
            return activity.__of__(self)
        raise KeyError(name)

    security.declarePrivate('objectIds')
    def objectIds(self, spec=None):
        return self._branches.keys()

    security.declarePrivate('objectValues')
    def objectValues(self, spec=None):
        return self._branches.values()

    security.declarePrivate('objectItems')
    def objectItems(self, spec=None):
        return self._branches.items()

InitializeClass(ZopeVersionHistory)





##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

from . import ZopeRepository
from AccessControl.class_init import InitializeClass
from App.ImageFile import ImageFile


def initialize(context):

    context.registerClass(
        instance_class = ZopeRepository.ZopeRepository,
        meta_type      = 'Repository',
        permission     = 'Add Repositories',
        constructors   = ZopeRepository.constructors,
        icon           = 'www/Repository.gif'
      )

    context.registerHelp()
    context.registerHelpTitle('Zope Help')

    registerIcon('VersionHistory.gif')
    registerIcon('Version.gif')


def install_hack():
    # Hackery - don't try this at home, kids. :) This is temporary for
    # testing purposes only.
    from .VersionSupport import VersionSupport
    import OFS.SimpleItem, App.Management

    method = App.Management.Tabs.filtered_manage_options
    def filtered_manage_options(self, REQUEST=None, method = method,
                                options = VersionSupport.manage_options):
        result = method(self, REQUEST)
        for item in result:
            if item.get('label') == 'Version Control':
                return result
        for option in options:
            result.append(option)
        return result
    App.Management.Tabs.filtered_manage_options = filtered_manage_options

    for _class in (OFS.SimpleItem.Item, OFS.SimpleItem.Item_w__name__):
        dict = _class.__dict__
        if not hasattr(dict, '__setitem__'):
            # new-style classes don't need this [>=2.8]
            continue

        for name, value in VersionSupport.__dict__.items():
            if name != 'manage_options':
                dict[name] = value

        InitializeClass(_class)


def registerIcon(filename):
    from OFS import misc_
    try:
        info = misc_.misc_.ZopeVersionControl
    except AttributeError:
        # Zope trunk
        return
    setattr(info, filename,
            ImageFile('www/%s' % filename, globals())
            )


zodbupdate_decode_dict = {
    'Products.ZopeVersionControl.EventLog LogEntry message': 'binary',
}



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

from . import Utility
from .EventLog import LogEntry
from .nonversioned import getNonVersionedData, restoreNonVersionedData
from .Utility import isAVersionableResource, VersionControlError, VersionInfo
from .Utility import use_vc_permission, _findPath
from .ZopeVersionHistory import ZopeVersionHistory
from AccessControl import ClassSecurityInfo
from Acquisition import Implicit, aq_parent, aq_inner
from AccessControl.class_init import InitializeClass
from BTrees.OIBTree import OIBTree
from BTrees.OOBTree import OOBTree
from DateTime.DateTime import DateTime
from Persistence import Persistent
from random import randint

import time


class Repository(Implicit, Persistent):
    """The repository implementation manages the actual data of versions
       and version histories. It does not handle user interface issues."""

    def __init__(self):
        # These keep track of symbolic label and branch names that
        # have been used to ensure that they don't collide.
        self._branches = OIBTree()
        self._branches['mainline'] = 1
        self._labels = OIBTree()

        self._histories = OOBTree()
        self._created = time.time()

    security = ClassSecurityInfo()

    security.declarePrivate('createVersionHistory')
    def createVersionHistory(self, object):
        """Internal: create a new version history for a resource."""
        # When one creates the first version in a version history, neither
        # the version or version history yet have a _p_jar, which causes
        # copy operations to fail. To work around that, we share our _p_jar.
        history_id = None
        while history_id is None or history_id in self._histories:
            history_id = str(randint(1, 9999999999))
        history = ZopeVersionHistory(history_id, object)
        self._histories[history_id] = history
        return history.__of__(self)

    security.declarePrivate('getVersionHistory')
    def getVersionHistory(self, history_id):
        """Internal: return a version history given a version history id."""
        return self._histories[history_id].__of__(self)

    security.declarePrivate('replaceState')
    def replaceState(self, obj, new_state):
        """Internal: replace the state of a persistent object.
        """
        non_versioned = getNonVersionedData(obj)
        # XXX There ought to be some way to do this more cleanly.
        # This fills the __dict__ of the old object with new state.
        # The other way to achieve the desired effect is to replace
        # the object in its container, but this method preserves the
        # identity of the object.
        if obj.__class__ is not new_state.__class__:
            raise VersionControlError(
                "The class of the versioned object has changed. %s != %s"
                % (repr(obj.__class__, new_state.__class__)))
        obj._p_changed = 1
        for key in list(obj.__dict__.keys()):
            if not key in new_state.__dict__:
                del obj.__dict__[key]
        for key, value in new_state.__dict__.items():
            obj.__dict__[key] = value
        if non_versioned:
            # Restore the non-versioned data into the new state.
            restoreNonVersionedData(obj, non_versioned)
        return obj

    #####################################################################
    # This is the implementation of the public version control interface.
    #####################################################################

    security.declarePublic('isAVersionableResource')
    def isAVersionableResource(self, obj):
        # For now, an object must be persistent (have its own db record)
        # in order to be considered a versionable resource.
        return isAVersionableResource(obj)

    security.declarePublic('isUnderVersionControl')
    def isUnderVersionControl(self, object):
        return hasattr(object, '__vc_info__')

    security.declarePublic('isResourceUpToDate')
    def isResourceUpToDate(self, object, require_branch=0):
        info = self.getVersionInfo(object)
        history = self.getVersionHistory(info.history_id)
        branch = 'mainline'
        if info.sticky:
            if info.sticky[0] == 'B':
                branch = info.sticky[1]
            elif require_branch:
                # The object is updated to a particular version
                # rather than a branch.  The caller
                # requires a branch.
                return 0
        return history.isLatestVersion(info.version_id, branch)

    security.declarePublic('isResourceChanged')
    def isResourceChanged(self, object):
        # Return true if the state of a resource has changed in a transaction
        # *after* the version bookkeeping was saved. Note that this method is
        # not appropriate for detecting changes within a transaction!
        info = self.getVersionInfo(object)
        itime = getattr(info, '_p_mtime', None)
        if itime is None:
            return 0
        mtime = Utility._findModificationTime(object)
        if mtime is None:
            return 0
        return mtime > itime

    security.declarePublic('getVersionInfo')
    def getVersionInfo(self, object):
        info = getattr(object, '__vc_info__', None)
        if info is not None:
            return info
        raise VersionControlError(
            'The specified resource is not under version control.'
            )

    security.declareProtected(use_vc_permission, 'applyVersionControl')
    def applyVersionControl(self, object, message=None):
        if self.isUnderVersionControl(object):
            raise VersionControlError(
                'The resource is already under version control.'
                )
        if not self.isAVersionableResource(object):
            raise VersionControlError(
                'This resource cannot be put under version control.'
                )

        # Need to check the parent to see if the container of the object
        # being put under version control is itself a version-controlled
        # object. If so, we need to use the branch id of the container.
        branch = 'mainline'
        parent = aq_parent(aq_inner(object))
        p_info = getattr(parent, '__vc_info__', None)
        if p_info is not None:
            sticky = p_info.sticky
            if sticky and sticky[0] == 'B':
                branch = sticky[1]

        # Create a new version history and initial version object.
        history = self.createVersionHistory(object)
        version = history.createVersion(object, branch)

        history_id = history.getId()
        version_id = version.getId()

        # Add bookkeeping information to the version controlled object.
        info = VersionInfo(history_id, version_id, VersionInfo.CHECKED_IN)
        if branch != 'mainline':
            info.sticky = ('B', branch)
        object.__vc_info__ = info

        # Save an audit record of the action being performed.
        history.addLogEntry(version_id,
                            LogEntry.ACTION_CHECKIN,
                            _findPath(object),
                            message is None and 'Initial checkin.' or message
                            )
        return object

    security.declareProtected(use_vc_permission, 'checkoutResource')
    def checkoutResource(self, object):
        info = self.getVersionInfo(object)
        if info.status != info.CHECKED_IN:
            raise VersionControlError(
                'The selected resource is already checked out.'
                )

        if info.sticky and info.sticky[0] != 'B':
            raise VersionControlError(
                'The selected resource has been updated to a particular '
                'version, label or date. The resource must be updated to '
                'the mainline or a branch before it may be checked out.'
                )

        if not self.isResourceUpToDate(object):
            raise VersionControlError(
                'The selected resource is not up to date!'
                )

        history = self.getVersionHistory(info.history_id)
        ob_path = _findPath(object)

        # Save an audit record of the action being performed.
        history.addLogEntry(info.version_id,
                            LogEntry.ACTION_CHECKOUT,
                            ob_path
                            )

        # Update bookkeeping information.
        newinfo = info.clone()
        newinfo.status = newinfo.CHECKED_OUT
        object.__vc_info__ = newinfo
        return object

    security.declareProtected(use_vc_permission, 'checkinResource')
    def checkinResource(self, object, message=''):
        info = self.getVersionInfo(object)
        if info.status != info.CHECKED_OUT:
            raise VersionControlError(
                'The selected resource is not checked out.'
                )

        if info.sticky and info.sticky[0] != 'B':
            raise VersionControlError(
                'The selected resource has been updated to a particular '
                'version, label or date. The resource must be updated to '
                'the mainline or a branch before it may be checked in.'
                )

        if not self.isResourceUpToDate(object):
            raise VersionControlError(
                'The selected resource is not up to date!'
                )

        history = self.getVersionHistory(info.history_id)
        ob_path = _findPath(object)

        branch = 'mainline'
        if info.sticky is not None and info.sticky[0] == 'B':
            branch = info.sticky[1]

        version = history.createVersion(object, branch)

        # Save an audit record of the action being performed.
        history.addLogEntry(version.getId(),
                            LogEntry.ACTION_CHECKIN,
                            ob_path,
                            message
                            )

        # Update bookkeeping information.
        newinfo = info.clone()
        newinfo.version_id = version.getId()
        newinfo.status = newinfo.CHECKED_IN
        object.__vc_info__ = newinfo
        return object

    security.declareProtected(use_vc_permission, 'uncheckoutResource')
    def uncheckoutResource(self, object):
        info = self.getVersionInfo(object)
        if info.status != info.CHECKED_OUT:
            raise VersionControlError(
                'The selected resource is not checked out.'
                )

        history = self.getVersionHistory(info.history_id)
        ob_path = _findPath(object)

        version = history.getVersionById(info.version_id)
        new_obj = version.copyState()

        # Save an audit record of the action being performed.
        history.addLogEntry(info.version_id,
                            LogEntry.ACTION_UNCHECKOUT,
                            ob_path
                            )

        # Replace the state of the object with a reverted state.
        new_obj = self.replaceState(object, new_obj)

        # Update bookkeeping information.
        newinfo = info.clone()
        newinfo.version_id = version.getId()
        newinfo.status = newinfo.CHECKED_IN
        new_obj.__vc_info__ = newinfo
        return new_obj

    security.declareProtected(use_vc_permission, 'updateResource')
    def updateResource(self, object, selector=None):
        info = self.getVersionInfo(object)
        if info.status != info.CHECKED_IN:
            raise VersionControlError(
                'The selected resource must be checked in to be updated.'
                )

        history = self.getVersionHistory(info.history_id)
        version = None
        sticky = info.sticky

        if not selector:
            # If selector is null, update to the latest version taking any
            # sticky attrs into account (branch, date). Note that the sticky
            # tag could also be a date or version id. We don't bother checking
            # for those, since in both cases we do nothing (because we'll
            # always be up to date until the sticky tag changes).
            if sticky and sticky[0] == 'L':
                # A label sticky tag, so update to that label (since it is
                # possible, but unlikely, that the label has been moved).
                version = history.getVersionByLabel(sticky[1])
            elif sticky and sticky[0] == 'B':
                # A branch sticky tag. Update to latest version on branch.
                version = history.getLatestVersion(selector)
            else:
                # Update to mainline, forgetting any date or version id
                # sticky tag that was previously associated with the object.
                version = history.getLatestVersion('mainline')
                sticky = None
        else:
            # If the selector is non-null, we find the version specified
            # and update the sticky tag. Later we'll check the version we
            # found and decide whether we really need to update the object.
            if type(selector) is str and history.hasVersionId(selector):
                version = history.getVersionById(selector)
                sticky = ('V', selector)

            elif type(selector) is str and selector in self._labels:
                version = history.getVersionByLabel(selector)
                sticky = ('L', selector)

            elif type(selector) is str and selector in self._branches:
                version = history.getLatestVersion(selector)
                if selector == 'mainline':
                    sticky = None
                else:
                    sticky = ('B', selector)
            else:
                try:
                    date = DateTime(selector)
                except Exception:
                    raise VersionControlError(
                        'Invalid version selector: %s' % selector
                        )
                else:
                    timestamp = date.timeTime()
                    sticky = ('D', timestamp)
                    # Fix!
                    branch = history.findBranchId(info.version_id)
                    version = history.getVersionByDate(branch, timestamp)

        # If the state of the resource really needs to be changed, do the
        # update and make a log entry for the update.
        version_id = version and version.getId() or info.version_id
        new_object = object
        if version and (version_id != info.version_id):
            new_object = version.copyState()
            new_object = self.replaceState(object, new_object)

            history.addLogEntry(version_id,
                                LogEntry.ACTION_UPDATE,
                                _findPath(new_object)
                                )

        # Update bookkeeping information.
        newinfo = info.clone(1)
        newinfo.version_id = version_id
        newinfo.status = newinfo.CHECKED_IN
        if sticky is not None:
            newinfo.sticky = sticky
        new_object.__vc_info__ = newinfo
        return new_object

    security.declareProtected(use_vc_permission, 'labelResource')
    def labelResource(self, object, label, force=0):
        info = self.getVersionInfo(object)
        if info.status != info.CHECKED_IN:
            raise VersionControlError(
                'The selected resource must be checked in to be labeled.'
                )

        # Make sure that labels and branch ids do not collide.
        if label in self._branches or label == 'mainline':
            raise VersionControlError(
                'The label value given is already in use as an activity id.'
                )
        if not label in self._labels:
            self._labels[label] = 1

        history = self.getVersionHistory(info.history_id)
        history.labelVersion(info.version_id, label, force)
        return object

    security.declareProtected(use_vc_permission, 'makeActivity')
    def makeActivity(self, object, branch_id):
        # Note - this is not part of the official version control API yet.
        # It is here to allow unit testing of the architectural aspects
        # that are already in place to support activities in the future.

        info = self.getVersionInfo(object)
        if info.status != info.CHECKED_IN:
            raise VersionControlError(
                'The selected resource must be checked in.'
                )

        branch_id = branch_id or None

        # Make sure that activity ids and labels do not collide.
        if branch_id in self._labels or branch_id == 'mainline':
            raise VersionControlError(
                'The value given is already in use as a version label.'
                )

        if not branch_id in self._branches:
            self._branches[branch_id] = 1

        history = self.getVersionHistory(info.history_id)

        if branch_id in history._branches:
            raise VersionControlError(
                'The resource is already associated with the given activity.'
                )

        history.createBranch(branch_id, info.version_id)
        return object

    security.declareProtected(use_vc_permission, 'getVersionOfResource')
    def getVersionOfResource(self, history_id, selector):
        history = self.getVersionHistory(history_id)
        sticky = None

        if not selector or selector == 'mainline':
            version = history.getLatestVersion('mainline')
        else:
            if history.hasVersionId(selector):
                version = history.getVersionById(selector)
                sticky = ('V', selector)

            elif selector in self._labels:
                version = history.getVersionByLabel(selector)
                sticky = ('L', selector)

            elif selector in self._branches:
                version = history.getLatestVersion(selector)
                sticky = ('B', selector)
            else:
                try: date = DateTime(selector)
                except:
                    raise VersionControlError(
                        'Invalid version selector: %s' % selector
                        )
                else:
                    timestamp = date.timeTime()
                    sticky = ('D', timestamp)
                    version = history.getVersionByDate('mainline', timestamp)

        object = version.copyState()

        info = VersionInfo(history_id, version.getId(), VersionInfo.CHECKED_IN)
        if sticky is not None:
            info.sticky = sticky
        object.__vc_info__ = info
        return object

    security.declareProtected(use_vc_permission, 'getVersionIds')
    def getVersionIds(self, object):
        info = self.getVersionInfo(object)
        history = self.getVersionHistory(info.history_id)
        return history.getVersionIds()

    security.declareProtected(use_vc_permission, 'getLabelsForResource')
    def getLabelsForResource(self, object):
        info = self.getVersionInfo(object)
        history = self.getVersionHistory(info.history_id)
        return history.getLabels()

    security.declareProtected(use_vc_permission, 'getLogEntries')
    def getLogEntries(self, object):
        info = self.getVersionInfo(object)
        history = self.getVersionHistory(info.history_id)
        return history.getLogEntries()

InitializeClass(Repository)



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################


class SequenceWrapper:
    """A helper that manages lazy acquisition wrapping."""
    def __init__(self, parent, items, pairs=None):
        if pairs is not None:
            self.pairs = 1
        self.parent = parent
        self.items = items

    pairs = None

    def __getitem__(self, key):
        item = self.items[key]
        if self.pairs is not None:
            return (item[0], item[1].__of__(self.parent))
        return item.__of__(self.parent)

    def __len__(self):
        return len(self.items)

    def __nonzero__(self):
        return len(self.items) > 0

